"""Tests for the Storage classes"""
import json
import sqlite3
import tempfile
from pathlib import Path
from typing import Any

import pytest
from veggie.storage import SQLiteStorage, Storage


@pytest.fixture
def temp_db_path() -> Any:
    """Creates a temporary SQLite database for testing."""
    with tempfile.NamedTemporaryFile(suffix=".sqlite") as temp_file:
        return temp_file.name


@pytest.fixture
def sqlite_storage(temp_db_path: str) -> Storage:
    """Returns an instance of SQLiteStorage with a temporary database."""
    return SQLiteStorage(path=temp_db_path)


def test_store_event_new(sqlite_storage: SQLiteStorage) -> None:
    """Test storing a new event."""
    event = {"uuid": "12345", "name": "task_completed", "sent_timestamp": 1670000000.0, "data": {"result": "success"}}
    sqlite_storage.store_event(event)

    with sqlite3.connect(sqlite_storage.path) as con:
        cursor = con.cursor()
        cursor.execute("SELECT data FROM events WHERE uuid = ?", (event["uuid"],))
        row = cursor.fetchone()
        assert row is not None
        assert json.loads(row[0]) == event


def test_store_event_update(sqlite_storage: SQLiteStorage) -> None:
    """Test updating an existing event."""
    event = {"uuid": "12345", "name": "task_started", "sent_timestamp": 1670000000.0, "data": {"status": "in_progress"}}
    sqlite_storage.store_event(event)

    updated_event = {
        "uuid": "12345",
        "name": "task_completed",
        "sent_timestamp": 1670000001.0,
        "data": {"status": "completed"},
    }
    sqlite_storage.store_event(updated_event)

    with sqlite3.connect(sqlite_storage.path) as con:
        cursor = con.cursor()
        cursor.execute("SELECT data FROM events WHERE uuid = ?", (event["uuid"],))
        row = cursor.fetchone()
        assert row is not None
        assert json.loads(row[0]) == updated_event


def test_get_events(sqlite_storage: SQLiteStorage) -> None:
    """Test retrieving all events."""
    events = [
        {"uuid": "1", "name": "task_1", "sent_timestamp": 1670000000.0, "data": {"result": "success"}},
        {"uuid": "2", "name": "task_2", "sent_timestamp": 1670000001.0, "data": {"result": "failure"}},
    ]
    for event in events:
        sqlite_storage.store_event(event)

    retrieved_events = sqlite_storage.get_events()
    assert len(retrieved_events) == len(events)
    assert retrieved_events[0]["uuid"] == "2"  # Check sorting by sent_timestamp


def test_get_event_by_id(sqlite_storage: SQLiteStorage) -> None:
    """Test retrieving a single event by its UUID."""
    event = {"uuid": "12345", "name": "task_completed", "sent_timestamp": 1670000000.0, "data": {"result": "success"}}
    sqlite_storage.store_event(event)

    retrieved_event = sqlite_storage.get_event_by_id("12345")
    assert retrieved_event == event


def test_get_event_by_id_not_found(sqlite_storage: SQLiteStorage) -> None:
    """Test retrieving a non-existent event by its UUID."""
    retrieved_event = sqlite_storage.get_event_by_id("non_existent_uuid")
    assert retrieved_event is None


def test_database_creation(sqlite_storage: SQLiteStorage) -> None:
    """Test if the database and table are created properly."""
    assert Path(sqlite_storage.path).exists()

    with sqlite3.connect(sqlite_storage.path) as con:
        cursor = con.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='events'")
        table = cursor.fetchone()
        assert table is not None


def test_index_creation(sqlite_storage: SQLiteStorage) -> None:
    """Test if the necessary indexes are created."""
    with sqlite3.connect(sqlite_storage.path) as con:
        cursor = con.cursor()
        cursor.execute("PRAGMA index_list(events)")
        indexes = cursor.fetchall()
        index_names = [index[1] for index in indexes]
        assert "idx_name" in index_names
        assert "idx_sent_timestamp" in index_names



"""
Celery monitor
"""
import asyncio
import os

from celery import Celery
from celery.events import EventReceiver
from kombu import Connection
from loguru import logger

from .storage import Storage
from .webapp.flask_app import get_flask_app


class CeleryMonitor:
    """
    Celery monitor class
    """

    def __init__(self, celery_app: Celery, storage: Storage) -> None:
        """
        Initializes the monitor
        """
        self.celery_app = celery_app
        self.storage = storage

        self.flask_app = get_flask_app(storage=storage, celery_app=celery_app)

    def _process_task_sent(self, event: dict) -> None:
        event["sent_timestamp"] = event["timestamp"]
        event.pop("timestamp")
        self.storage.store_event(event)

    def _process_task_received(self, event: dict) -> None:
        event["received_timestamp"] = event["timestamp"]
        event.pop("timestamp")
        self.storage.store_event(event)

    def _process_task_started(self, event: dict) -> None:
        event["started_timestamp"] = event["timestamp"]
        event.pop("timestamp")
        self.storage.store_event(event)

    def _process_task_succeeded(self, event: dict) -> None:
        event["succeeded_timestamp"] = event["timestamp"]
        event.pop("timestamp")
        self.storage.store_event(event)

    def _process_task_failed(self, event: dict) -> None:
        event["failed_timestamp"] = event["timestamp"]
        event.pop("timestamp")
        self.storage.store_event(event)

    async def start_event_receiver(self) -> None:
        """Starts the event receiver"""

        def _run_receiver() -> None:
            broker_url = self.celery_app.conf.broker_url
            with Connection(broker_url) as con:
                receiver = EventReceiver(
                    con,
                    handlers={
                        "task-sent": self._process_task_sent,
                        "task-received": self._process_task_received,
                        "task-started": self._process_task_started,
                        "task-succeeded": self._process_task_succeeded,
                        "task-failed": self._process_task_failed,
                    },
                )
                receiver.capture(limit=None, timeout=None)

        await asyncio.to_thread(_run_receiver)

    async def start_webapp(self) -> None:
        """Starts the webapp asynchronously"""

        def _run_webapp() -> None:
            """Starts the UI asynchronously"""
            self.flask_app.run(host="0.0.0.0", port=os.getenv("VEGGIE_PORT", 5000))

        await asyncio.to_thread(_run_webapp)

    def start(self) -> None:
        """Starts all services in the event loop"""

        async def _start_coroutine() -> None:
            event_receiver_task = asyncio.create_task(self.start_event_receiver())
            logger.info("Started Event Receiver")

            webapp_task = asyncio.create_task(self.start_webapp())
            logger.info("Started Web App")

            await asyncio.gather(event_receiver_task, webapp_task)

        asyncio.run(_start_coroutine())



""" Test celery monitor."""
from unittest.mock import Mock

import pytest
from celery import Celery
from veggie.monitor import CeleryMonitor
from veggie.storage import Storage


@pytest.fixture
def mock_storage() -> Mock:
    """Returns a mocked Storage instance."""
    return Mock(spec=Storage)


@pytest.fixture
def mock_celery_app() -> Mock:
    """Returns a mocked Celery instance."""
    celery_app = Mock(spec=Celery)
    celery_app.conf.broker_url = "redis://localhost:6379/0"
    return celery_app


@pytest.fixture
def celery_monitor(mock_celery_app: Mock, mock_storage: Mock) -> CeleryMonitor:
    """Returns an instance of CeleryMonitor."""
    return CeleryMonitor(mock_celery_app, mock_storage)


def test_process_task_sent(celery_monitor: CeleryMonitor, mock_storage: Mock) -> None:
    """Test processing of 'task-sent' event."""
    event = {"uuid": "123", "timestamp": 1670000000.0, "name": "task_sent"}
    celery_monitor._process_task_sent(event)

    assert "timestamp" not in event
    assert event["sent_timestamp"] == 1670000000.0
    mock_storage.store_event.assert_called_once_with(event)


def test_process_task_received(celery_monitor: CeleryMonitor, mock_storage: Mock) -> None:
    """Test processing of 'task-received' event."""
    event = {"uuid": "123", "timestamp": 1670000000.0, "name": "task_received"}
    celery_monitor._process_task_received(event)

    assert "timestamp" not in event
    assert event["received_timestamp"] == 1670000000.0
    mock_storage.store_event.assert_called_once_with(event)


def test_process_task_started(celery_monitor: CeleryMonitor, mock_storage: Mock) -> None:
    """Test processing of 'task-started' event."""
    event = {"uuid": "123", "timestamp": 1670000000.0, "name": "task_started"}
    celery_monitor._process_task_started(event)

    assert "timestamp" not in event
    assert event["started_timestamp"] == 1670000000.0
    mock_storage.store_event.assert_called_once_with(event)


def test_process_task_succeeded(celery_monitor: CeleryMonitor, mock_storage: Mock) -> None:
    """Test processing of 'task-succeeded' event."""
    event = {"uuid": "123", "timestamp": 1670000000.0, "name": "task_succeeded"}
    celery_monitor._process_task_succeeded(event)

    assert "timestamp" not in event
    assert event["succeeded_timestamp"] == 1670000000.0
    mock_storage.store_event.assert_called_once_with(event)


def test_process_task_failed(celery_monitor: CeleryMonitor, mock_storage: Mock) -> None:
    """Test processing of 'task-failed' event."""
    event = {"uuid": "123", "timestamp": 1670000000.0, "name": "task_failed"}
    celery_monitor._process_task_failed(event)

    assert "timestamp" not in event
    assert event["failed_timestamp"] == 1670000000.0
    mock_storage.store_event.assert_called_once_with(event)



"""Storage backend for Celery Admin"""
import json
import pathlib
import sqlite3
from abc import ABC, abstractmethod
from typing import Any


class Storage(ABC):
    """Abstract Storage class"""

    @abstractmethod
    def store_event(self, event: dict) -> None:
        """Method to store event info"""
        pass

    @abstractmethod
    def get_events(self) -> list[dict]:
        """Method to get events"""
        pass

    @abstractmethod
    def get_event_by_id(self, id: str) -> dict | None:
        """Method to get single event by its ID"""
        pass


class SQLiteStorage(Storage):
    """SQLite Backend"""

    def __init__(self, path: str) -> None:
        """Initializes SQLite Storage"""
        self.path = str(pathlib.Path(path).resolve())
        self.table_name = "events"

        with self._get_connection() as con:
            cursor = con.cursor()
            cursor.execute(
                f"""
                    CREATE TABLE IF NOT EXISTS {self.table_name} (
                        uuid TEXT,
                        data JSON,
                        name TEXT GENERATED ALWAYS AS (json_extract(data, '$.name')) VIRTUAL,
                        sent_timestamp REAL GENERATED ALWAYS AS (json_extract(data, '$.sent_timestamp')) VIRTUAL,
                    PRIMARY KEY (uuid)
                    )
                """
            )
            cursor.execute(f"CREATE INDEX IF NOT EXISTS idx_name ON {self.table_name} (name)")
            cursor.execute(f"CREATE INDEX IF NOT EXISTS idx_sent_timestamp ON {self.table_name} (sent_timestamp)")

    def _get_connection(self) -> Any:
        return sqlite3.connect(self.path)

    def store_event(self, event: dict) -> None:
        """Stores event info"""
        with self._get_connection() as con:
            cursor = con.cursor()
            cursor.execute(f"SELECT data FROM {self.table_name} WHERE uuid = ?", (event["uuid"],))
            row = cursor.fetchone()

            if row:
                # Record exists, merge existing data with new data
                data = json.loads(row[0])
                data.update(event)
                cursor.execute(
                    f"UPDATE {self.table_name} SET data = ? WHERE uuid = ?", (json.dumps(data), event["uuid"])
                )
            else:
                # Record does not exist, insert a new one
                cursor.execute(
                    f"INSERT INTO {self.table_name} (uuid, data) VALUES (?, ?)", (event["uuid"], json.dumps(event))
                )

    def get_events(self) -> list[dict]:
        """Gets events as list of dicts"""
        with self._get_connection() as con:
            cursor = con.cursor()
            cursor.execute(f"SELECT data FROM {self.table_name} ORDER BY sent_timestamp DESC")
            rows = cursor.fetchall()
            return [json.loads(row[0]) for row in rows]

    def get_event_by_id(self, id: str) -> dict | None:
        """Gets single event by its ID"""
        with self._get_connection() as con:
            cursor = con.cursor()
            cursor.execute(f"SELECT data FROM {self.table_name} WHERE uuid = ?", (id,))
            row = cursor.fetchone()
            if row is None:
                return None
            return json.loads(row[0])
